import os
import sys
import math
import random
import numpy as np
import os.path as osp
from datetime import datetime

import torch
import torch.nn as nn
import torch.optim as optim
import torch.autograd as autograd
import torch.utils.data as tordata

from .tripletloss import TripletLoss
from .tripletsampler import TripletSampler


class Model:
    def __init__(self, hidden_dim, lr, hard_or_full_trip, margin, num_workers, frame_num, batch_size, restore_iter, total_iter, save_iter, print_iter, 
                 save_name, model_name, data_source, is_train, img_size):
        self.model_name = model_name  
        self.save_name = save_name
        self.restore_iter = restore_iter
        self.total_iter = total_iter   
        self.save_iter = save_iter
        self.print_iter = print_iter

        self.is_train = is_train
        self.data_source = data_source
        self.num_workers = num_workers
        self.frame_num = frame_num
               
        self.batch_size = batch_size
        self.P, self.M = batch_size  
        self.lr = lr
        self.hidden_dim = hidden_dim
        self.hard_or_full_trip = hard_or_full_trip
        self.margin = margin

        # GaitSetNet
        if self.model_name == 'GaitSetNet':
            from .net_GaitSetNet import GaitSetNet as SetNet
            print('     Model Info: name={} frameNum={}'.format(self.model_name, self.frame_num))
        # GaitMSNet           
        if self.model_name == 'GaitLGNet': 
            from .net_GaitCOMPNet import SlowFastNet as SetNet
            print('     Model Info: name={} frameNum={}'.format(self.model_name, self.frame_num))
        self.encoder = SetNet(self.hidden_dim).float()
        self.encoder = nn.DataParallel(self.encoder)
        self.encoder.cuda()
 
        self.triplet_label_loss = TripletLoss(self.P * self.M, self.hard_or_full_trip, self.margin).float()
        self.triplet_label_loss = nn.DataParallel(self.triplet_label_loss)
        self.triplet_label_loss.cuda()

        self.loss_metric = list()
        self.dist_list , self.hard_loss_metric, self.full_loss_metric, self.full_loss_num = list(), list(), list(), list()
        self.optimizer = optim.Adam([{'params': self.encoder.parameters()}], lr=self.lr)

        self.sample_type = 'all' 

    def save(self):
        os.makedirs(osp.join('checkpoint', self.model_name), exist_ok=True)
        torch.save(self.encoder.state_dict(), osp.join('checkpoint', self.model_name, '{}-{:0>5}-encoder.ptm'.format(self.save_name, self.restore_iter)))
        torch.save(self.optimizer.state_dict(), osp.join('checkpoint', self.model_name, '{}-optimizer.ptm'.format(self.save_name)))
                                
    def load(self, restore_iter):
        self.encoder.load_state_dict(torch.load(osp.join('checkpoint', self.model_name, '{}-{:0>5}-encoder.ptm'.format(self.save_name, restore_iter))))
        self.optimizer.load_state_dict(torch.load(osp.join('checkpoint', self.model_name, '{}-optimizer.ptm'.format(self.save_name))))
            
    def collate_fn(self, batch):
        batch_size = len(batch)
        feature_num = len(batch[0][0])
        seqs_fast = [batch[i][0] for i in range(batch_size)]
        seqs_slow = [batch[i][1] for i in range(batch_size)]
        framesets_fast = [batch[i][2] for i in range(batch_size)]
        framesets_slow = [batch[i][3] for i in range(batch_size)]
        seqs_view = [batch[i][4] for i in range(batch_size)]
        seqs_type = [batch[i][5] for i in range(batch_size)]
        seqs_label = [batch[i][6] for i in range(batch_size)]
        batch = [seqs_slow, seqs_fast, seqs_view, seqs_type, seqs_label, None, None]

        def select_slowframe(index):
            sample = seqs_slow[index]
            frame_set = framesets_slow[index]
            if self.sample_type == 'slowfast':
                frame_id_list = random.choices(frame_set, k=self.frame_num[1])
                _ = [feature.loc[frame_id_list].values for feature in sample]
            else:
                _ = [feature.values for feature in sample]
            return _
        seqs_slow = list(map(select_slowframe, range(len(seqs_slow))))

        def select_fastframe(index):
            sample = seqs_fast[index]
            frame_set = framesets_fast[index]
            if self.sample_type == 'slowfast':
                frame_id_list = np.asarray(range(self.frame_num[0])) + random.randint(0, max(0, len(frame_set) - self.frame_num[0]))
                for j in range(self.frame_num[0]):
                    while frame_id_list[j] >= len(frame_set): frame_id_list[j] = frame_id_list[j] - len(frame_set) 
                _ = [feature.loc[frame_id_list].values for feature in sample]
            else:
                _ = [feature.values for feature in sample]
            return _
        seqs_fast = list(map(select_fastframe, range(len(seqs_fast))))

        if self.sample_type == 'slowfast':
            seqs_slow = [np.asarray([seqs_slow[i][j] for i in range(batch_size)]) for j in range(feature_num)]
            seqs_fast = [np.asarray([seqs_fast[i][j] for i in range(batch_size)]) for j in range(feature_num)]
        else:
            # seqs_slow
            gpu_num = min(torch.cuda.device_count(), batch_size)
            batch_per_gpu = math.ceil(batch_size / gpu_num)
            batch_frames = [[len(framesets_slow[i]) for i in range(batch_per_gpu * _, batch_per_gpu * (_ + 1)) if i < batch_size] for _ in range(gpu_num)]
            if len(batch_frames[-1]) != batch_per_gpu:
                for _ in range(batch_per_gpu - len(batch_frames[-1])): batch_frames[-1].append(0)
            max_sum_frame = np.max([np.sum(batch_frames[_]) for _ in range(gpu_num)])
            seqs_slow = [[np.concatenate([seqs_slow[i][j] for i in range(batch_per_gpu * _, batch_per_gpu * (_ + 1)) if i < batch_size], 0) for _ in range(gpu_num)]for j in range(feature_num)]
            seqs_slow = [np.asarray([np.pad(seqs_slow[j][_], ((0, max_sum_frame - seqs_slow[j][_].shape[0]), (0, 0), (0, 0)), 'constant', constant_values=0) for _ in range(gpu_num)]) for j in range(feature_num)]
            batch[5] = np.asarray(batch_frames)
            # seqs_fast
            gpu_num = min(torch.cuda.device_count(), batch_size)
            batch_per_gpu = math.ceil(batch_size / gpu_num)
            batch_frames = [[len(framesets_fast[i]) for i in range(batch_per_gpu * _, batch_per_gpu * (_ + 1)) if i < batch_size] for _ in range(gpu_num)]
            if len(batch_frames[-1]) != batch_per_gpu:
                for _ in range(batch_per_gpu - len(batch_frames[-1])): batch_frames[-1].append(0)
            max_sum_frame = np.max([np.sum(batch_frames[_]) for _ in range(gpu_num)])
            seqs_fast = [[np.concatenate([seqs_fast[i][j] for i in range(batch_per_gpu * _, batch_per_gpu * (_ + 1)) if i < batch_size], 0) for _ in range(gpu_num)]for j in range(feature_num)]
            seqs_fast = [np.asarray([np.pad(seqs_fast[j][_], ((0, max_sum_frame - seqs_fast[j][_].shape[0]), (0, 0), (0, 0)), 'constant', constant_values=0) for _ in range(gpu_num)]) for j in range(feature_num)]
            batch[6] = np.asarray(batch_frames)

        batch[0] = seqs_slow
        batch[1] = seqs_fast
        return batch

    def fit(self):
        if self.restore_iter != 0: self.load(self.restore_iter)

        self.encoder.train()
        self.sample_type = 'slowfast'   
        for param_group in self.optimizer.param_groups: param_group['lr'] = self.lr

        train_label_set = sorted(list(self.data_source.seqs_label_set)) 
        triplet_sampler = TripletSampler(self.data_source, self.batch_size)
        train_loader = tordata.DataLoader(dataset=self.data_source, batch_sampler=triplet_sampler, collate_fn=self.collate_fn, num_workers=self.num_workers)

        _time = datetime.now()
        for batch in train_loader:
            self.restore_iter += 1
            self.optimizer.zero_grad()
            
            seqs_slow, seqs_fast, seqs_view, seqs_type, seqs_label, batchframe_slow, batchframe_fast = batch[0], batch[1], batch[2], batch[3], batch[4], batch[5], batch[6]
            for i in range(len(seqs_slow)): seqs_slow[i] = self.np2var(seqs_slow[i]).float()
            for i in range(len(seqs_fast)): seqs_fast[i] = self.np2var(seqs_fast[i]).float()
            if batchframe_slow is not None: batchframe_slow = self.np2var(batchframe_slow).int()
            if batchframe_fast is not None: batchframe_fast = self.np2var(batchframe_fast).int()
            feat_id, _ = self.encoder(*seqs_slow, *seqs_fast, batchframe_slow, batchframe_fast)
            target_label = [train_label_set.index(l) for l in seqs_label]
            target_label = self.np2var(np.array(target_label)).long()
            triplet_feature = feat_id.permute(1, 0, 2).contiguous()
            triplet_label = target_label.unsqueeze(0).repeat(triplet_feature.size(0), 1)
            full_loss_metric, hard_loss_metric, mean_dist, full_loss_num = self.triplet_label_loss(triplet_feature, triplet_label)
            if self.hard_or_full_trip == 'hard': loss = hard_loss_metric.mean()
            elif self.hard_or_full_trip == 'full': loss = full_loss_metric.mean()
            self.loss_metric.append(loss.mean().data.cpu().numpy())         
            self.hard_loss_metric.append(hard_loss_metric.mean().data.cpu().numpy())
            self.full_loss_metric.append(full_loss_metric.mean().data.cpu().numpy())
            self.full_loss_num.append(full_loss_num.mean().data.cpu().numpy())
            self.dist_list.append(mean_dist.mean().data.cpu().numpy())
                
            if loss > 1e-9:
                loss.backward()
                self.optimizer.step()

            if self.restore_iter % self.print_iter == 0:
                print('     iter-{}:'.format(self.restore_iter), end='')
                print(', loss={0:.8f}'.format(np.mean(self.loss_metric)), end='')
                print(', h_loss_metric={0:.8f}'.format(np.mean(self.hard_loss_metric)), end='')
                print(', f_loss_metric={0:.8f}'.format(np.mean(self.full_loss_metric)), end='')
                print(', f_loss_num={0:.8f}'.format(np.mean(self.full_loss_num)), end='')
                print(', mean_dist={0:.8f}'.format(np.mean(self.dist_list)), end='')
                print(', lr={0:.8f}'.format(self.lr), end='\n')
                sys.stdout.flush()
                self.loss_metric = list()
                self.dist_list , self.hard_loss_metric, self.full_loss_metric, self.full_loss_num = list(), list(), list(), list()

            if self.restore_iter % self.save_iter == 0: 
                self.save()
                print('-----@ {} total training cost={}'.format(datetime.now(), (datetime.now()-_time)), end='\n')

            if self.restore_iter % 30000 == 0: 
                self.lr = self.lr * 0.1
                for param_group in self.optimizer.param_groups: param_group['lr'] = self.lr             
                   
            if self.restore_iter == self.total_iter:
                break

    def ts2var(self, x):
        return autograd.Variable(x).cuda()

    def np2var(self, x):
        return self.ts2var(torch.from_numpy(x))

    def transform(self, batch_size=1):
        assert(not self.is_train)
        self.encoder.eval()
        self.sample_type = 'all'  
        data_loader = tordata.DataLoader(dataset=self.data_source, batch_size=batch_size, sampler=tordata.sampler.SequentialSampler(self.data_source), collate_fn=self.collate_fn, num_workers=self.num_workers)

        list_feature = list()
        list_view, list_type, list_label = list(), list(), list()
        for i, x in enumerate(data_loader):
            seq_slow, seq_fast, seq_view, seq_type, seq_label, batchframe_slow, batchframe_fast = x
            for j in range(len(seq_slow)): seq_slow[j] = self.np2var(seq_slow[j]).float()
            for j in range(len(seq_fast)): seq_fast[j] = self.np2var(seq_fast[j]).float()
            if batchframe_slow is not None: batchframe_slow = self.np2var(batchframe_slow).int()
            if batchframe_fast is not None: batchframe_fast = self.np2var(batchframe_fast).int()
            feature, _ = self.encoder(*seq_slow, *seq_fast, batchframe_slow, batchframe_fast)
            n, num_bin, _ = feature.size()
            list_feature.append(feature.view(n, -1).data.cpu().numpy())
            list_view += seq_view
            list_type += seq_type
            list_label += seq_label
        return np.concatenate(list_feature, 0), list_view, list_type, list_label
