import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np

BN_MOMENTUM = 0.1


class HRModule(nn.Module):
    def __init__(self, num_branches, num_channels):
        super(HRModule, self).__init__()
        self.num_branches = num_branches
        self.num_channels = num_channels
        self.relu = nn.ReLU(True)
        self.branches = self._make_branches()
        self.fuse_layers = self._make_fuse_layers()

    def _make_branches(self):
        branches = []
        for i in range(self.num_branches):
            branches.append(nn.Sequential(
                nn.Conv2d(self.num_channels[i], self.num_channels[i], 1, 1, 0, bias=False),
                nn.BatchNorm2d(self.num_channels[i], momentum=BN_MOMENTUM),
                nn.ReLU(inplace=True)))
        return nn.ModuleList(branches)

    def _make_fuse_layers(self):
        fuse_layers = []
        for i in range(self.num_branches):
            fuse_layer = []
            for j in range(self.num_branches):
                if j == i:
                    fuse_layer.append(None)
                elif j > i:
                    fuse_layer.append(nn.Sequential(
                        nn.Conv2d(self.num_channels[j], self.num_channels[i], 1, 1, 0, bias=False),
                        nn.BatchNorm2d(self.num_channels[i]),
                        nn.Upsample(scale_factor=2**(j-i), mode='nearest')))
                else:
                    conv3x3s = []
                    for k in range(i-j):
                        if k < i - j - 1:
                            conv3x3s.append(nn.Sequential(
                                nn.Conv2d(self.num_channels[j], self.num_channels[j], 3, 2, 1, bias=False),
                                nn.BatchNorm2d(self.num_channels[j]),
                                nn.ReLU(True)))
                        else:
                            conv3x3s.append(nn.Sequential(
                                nn.Conv2d(self.num_channels[j], self.num_channels[i], 3, 2, 1, bias=False),
                                nn.BatchNorm2d(self.num_channels[i])))
                    fuse_layer.append(nn.Sequential(*conv3x3s))
            fuse_layers.append(nn.ModuleList(fuse_layer))
        return nn.ModuleList(fuse_layers)

    def forward(self, x):
        for i in range(self.num_branches):
            x[i] = self.branches[i](x[i])

        x_fuse = []
        for i in range(len(self.fuse_layers)):
            y = x[0] if i == 0 else self.fuse_layers[i][0](x[0])
            for j in range(1, self.num_branches):
                if i == j:
                    y = y + x[j]
                else:
                    y = y + self.fuse_layers[i][j](x[j])
            x_fuse.append(self.relu(y))
        return x_fuse


class MSNet(nn.Module):
    def __init__(self, hidden_dim):
        super(MSNet, self).__init__()
        self.hidden_dim = hidden_dim
        self.batch_frame = None

        # Stage1 (downsample)
        self.conv1 = nn.Conv2d(1, 64, 3, 2, 1, bias=False)
        self.bn1 = nn.BatchNorm2d(64, momentum=BN_MOMENTUM)
        self.conv2 = nn.Conv2d(64, 64, 3, 1, 1, bias=False)
        self.bn2 = nn.BatchNorm2d(64, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        # Stage2 (expand branch and fuse branch)
        self.stage2_param = {'BRANCHES': 2, 'CHANNELS': [64, 128],}
        self.expansion1 = self._expand_branch([64], self.stage2_param['CHANNELS'])
        self.stage2 = self._fuse_branch(self.stage2_param)
        # Stage3 (expand branch and fuse branch)
        self.stage3_param = {'BRANCHES': 3, 'CHANNELS': [64, 128, 256],}
        self.expansion2 = self._expand_branch(self.stage2_param['CHANNELS'], self.stage3_param['CHANNELS'])
        self.stage3 = self._fuse_branch(self.stage3_param)
        # Stage4 (map branch)
        self.mapping = self._map_branch(self.stage3_param)
        # HPM x[0]
        self.bin_num1 = [1, 2, 3, 6, 12]
        self.fc_bin1 = nn.ParameterList([nn.Parameter(nn.init.xavier_uniform_(torch.zeros(sum(self.bin_num1), 64, self.hidden_dim)))])      
        # HPM x[1]
        self.bin_num2 = [1, 2, 3, 6]
        self.fc_bin2 = nn.ParameterList([nn.Parameter(nn.init.xavier_uniform_(torch.zeros(sum(self.bin_num2), 128, self.hidden_dim)))])

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.xavier_uniform_(m.weight.data)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight.data)
                nn.init.constant_(m.bias.data, 0.0)

    def _expand_branch(self, channels_pre_layer, channels_cur_layer):
        expandsion_layers = []
        for i in range(len(channels_cur_layer)):
            if i < len(channels_pre_layer):
                if channels_cur_layer[i] == channels_pre_layer[i]:
                    expandsion_layers.append(None)
                else:
                    expandsion_layers.append(nn.Sequential(
                        nn.Conv2d(channels_pre_layer[i], channels_cur_layer[i], 3, 1, 1, bias=False),
                        nn.BatchNorm2d(channels_cur_layer[i]),
                        nn.ReLU(inplace=True)))
            else:
                expandsion_layers.append(nn.Sequential(
                    nn.Conv2d(channels_pre_layer[-1], channels_cur_layer[i], 3, 2, 1, bias=False),
                    nn.BatchNorm2d(channels_cur_layer[i]),
                    nn.ReLU(inplace=True)))
        return nn.ModuleList(expandsion_layers)

    def _fuse_branch(self, layer_config):
        num_branches = layer_config['BRANCHES']
        num_channels = layer_config['CHANNELS']
        module = HRModule(num_branches, num_channels)
        return module

    def _map_branch(self, layer_config):
        mapping_layers = []
        for i in range(layer_config['BRANCHES']): 
            mapping_layers.append(nn.Sequential(
                nn.Conv2d(layer_config['CHANNELS'][i], layer_config['CHANNELS'][i], 3, 1, 1, bias=False),
                nn.BatchNorm2d(layer_config['CHANNELS'][i]),
                nn.ReLU(inplace=True)))                
        return nn.ModuleList(mapping_layers)

    def frame_max(self, x):
        if self.batch_frame is None:
            return torch.max(x, 1)
        else:
            _tmp = [torch.max(x[:, self.batch_frame[i]:self.batch_frame[i + 1], :, :, :], 1) for i in range(len(self.batch_frame) - 1)]
            max_list = torch.cat([_tmp[i][0] for i in range(len(_tmp))], 0)
            arg_max_list = torch.cat([_tmp[i][1] for i in range(len(_tmp))], 0)
            return max_list, arg_max_list

    def forward(self, silho, batch_frame=None):
        # n: batch_size, s: frame_num, k: keypoints_num, c: channel
        if batch_frame is not None:
            batch_frame = batch_frame[0].data.cpu().numpy().tolist()
            _ = len(batch_frame)
            for i in range(len(batch_frame)):
                if batch_frame[-(i + 1)] != 0:
                    break
                else:
                    _ -= 1
            batch_frame = batch_frame[:_]
            frame_sum = np.sum(batch_frame)
            if frame_sum < silho.size(1): silho = silho[:, :frame_sum, :, :]
            self.batch_frame = [0] + np.cumsum(batch_frame).tolist()
        n = silho.size(0)
        x = silho.unsqueeze(2)
        del silho
        x = x[:, :, :, 40:, :]

        # reshape Tensor (5D -> 4D)
        n, s, c, h, w = x.size()
        x = x.view(-1, c, h, w)
        # Stage1 (downsample)
        x = self.bn1(self.conv1(x))
        x = self.bn2(self.conv2(x))
        x = self.relu(x)
        # Stage2 (expand branch and fuse branch)
        x_list = []
        for i in range(self.stage2_param['BRANCHES']):
            if self.expansion1[i] is None:
                x_list.append(x)
            else: 
                x_list.append(self.expansion1[i](x))
        y_list = self.stage2(x_list)
        # Stage3 (expand branch and fuse branch)
        x_list = []
        for i in range(self.stage3_param['BRANCHES']):
            if self.expansion2[i] is None:
                x_list.append(y_list[i])
            else: 
                x_list.append(self.expansion2[i](y_list[-1]))
        y_list = self.stage3(x_list)
        # Stage4 (map branch)
        x_list = []
        for i in range(self.stage3_param['BRANCHES']):
            x_list.append(self.mapping[i](y_list[i]))
        # reshape Tensor (4D -> 5D)
        s_list = []
        for x in x_list: 
            _, c, h, w = x.size()
            x = x.view(n, s, c, h, w)
            s_list.append(self.frame_max(x)[0])
         
        feature1, feature2 = list(), list()
        n, c, _, _ = s_list[0].size()
        for num_bin in self.bin_num1:
            z = s_list[0].view(n, c, num_bin, -1)
            z = z.mean(3) + z.max(3)[0]
            feature1.append(z)
        feature1 = torch.cat(feature1, 2).permute(2, 0, 1).contiguous()
        feature1 = feature1.matmul(self.fc_bin1[0])
        feature1 = feature1.permute(1, 0, 2).contiguous()
        n, c, _, _ = s_list[1].size()
        for num_bin in self.bin_num2:
            z = s_list[1].view(n, c, num_bin, -1)
            z = z.mean(3) + z.max(3)[0]
            feature2.append(z)
        feature2 = torch.cat(feature2, 2).permute(2, 0, 1).contiguous()
        feature2 = feature2.matmul(self.fc_bin2[0])
        feature2 = feature2.permute(1, 0, 2).contiguous()
        
        feat_id = torch.cat([feature1, feature2], dim=1)
        return feat_id


