import argparse
import os
import shutil
import itertools
import numpy as np
import xarray as xr
from datetime import datetime

import torch
import torch.nn.functional as F

from config import conf as config
from initialization import initialization

times = 1
num_rank = 5

# Compute euclidean distance
def cuda_dist(x, y):
    x = torch.from_numpy(x).cuda()
    y = torch.from_numpy(y).cuda()
    dist = torch.sum(x ** 2, 1).unsqueeze(1) + torch.sum(y ** 2, 1).unsqueeze(1).transpose(0, 1) - 2 * torch.matmul(x, y.transpose(0, 1))
    dist = torch.sqrt(F.relu(dist))
    return dist
    
def evaluation(data, config):
    dataset = config['dataset_name'].split('-')[0]
    probe_seq_dict = {'CASIA': [['nm-04', 'nm-05', 'nm-06'], ['cl-02']]}
    gallery_seq_dict = {'CASIA': [['nm-01', 'nm-02', 'nm-03', 'cl-01']]}

    feature, list_view, list_type, list_label = data
    label = np.array(list_label)
    view_list = sorted(list(set(list_view)))
    view_num, sample_num = len(view_list), len(feature)
    acc = np.zeros([len(probe_seq_dict[dataset]), view_num, num_rank])
    for p, probe_seq in enumerate(probe_seq_dict[dataset]):
        for gallery_seq in gallery_seq_dict[dataset]:
            for v, view in enumerate(view_list):
                gseq_mask = np.isin(list_type, gallery_seq) & np.isin(list_view, [view])
                gallery_x, gallery_y = feature[gseq_mask, :], label[gseq_mask]
                pseq_mask = np.isin(list_type, probe_seq) & np.isin(list_view, [view])
                probe_x, probe_y = feature[pseq_mask, :], label[pseq_mask]
                dist = cuda_dist(probe_x, gallery_x)
                idx = dist.sort(1)[1].cpu().numpy()
                acc[p, v, :] = np.round(np.sum(np.cumsum(np.reshape(probe_y, [-1, 1]) == gallery_y[idx[:, 0:num_rank]], 1) > 0, 0) * 100 / dist.shape[0], 2)
    return acc


for _time in range(2, 9):
    # Set Paramater
    config['data']['view_angle'] = str(_time * 18).zfill(3)
	
    # Train
    print(' ==> Training %d START' % _time)
    is_train = True
    parser = argparse.ArgumentParser(description='Train')
    parser.add_argument('--cache', default=False, type=bool, help='cache: if set as TRUE all the training data will be loaded at once before the training start. Default: TRUE')
    opt = parser.parse_args()
    model = initialization(config, cache=opt.cache, is_train=is_train)
    model.fit()
    print(' ==> Training COMPLETE\n')

    # Test
    print(' ==> Testing %d START' % _time)
    is_train = False
    parser = argparse.ArgumentParser(description='Test')
    parser.add_argument('--batch_size', default='1', type=int, help='batch_size: batch size for parallel test. Default: 1')
    parser.add_argument('--cache', default=False, type=bool, help='cache: if set as TRUE all the test data will be loaded at once before the transforming start. Default: FALSE')
    opt = parser.parse_args()
    model = initialization(config, cache=opt.cache, is_train=is_train)
    best_iter, best_acc = 0, 0
    for iter in range(config['model']['save_iter'], config['model']['total_iter']+1, config['model']['save_iter']):
        print(' ==> Loading the model of iteration %d...' % iter)
        np.set_printoptions(precision=2, floatmode='fixed')
        model.load(iter)
        time = datetime.now()
        for i in range(times):
            data = model.transform(opt.batch_size)
            acc_ = evaluation(data, config['data'])
            acc = acc_ if i == 0 else acc+acc_
        acc = acc / times           
        print('Evaluation complete. Cost:', datetime.now() - time)
        print('***** Rank-1 %s NM: %.3f \t CL: %.3f *****' % (config['data']['view_angle'], np.mean(acc[0, :, 0]), np.mean(acc[1, :, 0])))
        ave_acc = (np.mean(acc[0, :, 0]) + np.mean(acc[1, :, 0])) / 2
        if ave_acc >= best_acc: best_iter, best_acc = iter, ave_acc
        print('***** best_iter: %d\t best_acc: %0.3f\t ave_acc: %0.3f' % (best_iter, best_acc, ave_acc))
    print(' ==> Testing FINISH\n')

    # Save
    print(' ==> Recoding')
    print(' ==> Loading the model of best iteration %d...' % best_iter)
    model.load(best_iter)
    time = datetime.now()
    np.set_printoptions(precision=2, floatmode='fixed')
    time = datetime.now()
    data = model.transform(opt.batch_size)
    for i in range(times):
        data = model.transform(opt.batch_size)
        acc_ = evaluation(data, config['data'])
        acc = acc_ if i == 0 else acc+acc_
    acc = acc / times           
    print('Evaluation complete. Cost:', datetime.now() - time)
    print('***** Rank-1 NM: %.3f \t CL: %.3f *****' % (np.mean(acc[0, :, 0]), np.mean(acc[1, :, 0])))
    with open('RECORD', 'a') as fout:
        fout.write('model_name: {}\t view: {}\t batch_size: {}\t frame_num: {}'.format(config['model']['model_name'], config['data']['view_angle'], config['model']['batch_size'], config['model']['frame_num']))
        fout.write('Rank-1 : NM: %.3f \t CL: %.3f\n' % (np.mean(acc[0, :,0]), np.mean(acc[1, :,0])))
    print(' ==> Recoding FINISH\n\n')
