import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np


class SetBlock(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, pooling=False, **kwargs):
        super(SetBlock, self).__init__()
        self.pooling = pooling
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, bias=False, **kwargs)
        if pooling: self.pool2d = nn.MaxPool2d(2)
    def forward(self, x):
        n, s, c, h, w = x.size()
        x = F.leaky_relu(self.conv(x.view(-1,c,h,w)), inplace=True)
        if self.pooling: x = self.pool2d(x)
        _, c, h, w = x.size()
        return x.view(n, s, c, h ,w)


class PipeBlock(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, pooling=False, **kwargs):
        super(PipeBlock, self).__init__()
        self.pooling = pooling
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, bias=False, **kwargs)
        if pooling: self.pool2d = nn.MaxPool2d(2)
    def forward(self, x):
        x = self.conv(x)
        x = F.leaky_relu(x, inplace=True)
        if self.pooling: x = self.pool2d(x)
        return x


class FSNet(nn.Module):
    def __init__(self, hidden_dim):
        super(FSNet, self).__init__()
        self.hidden_dim = hidden_dim
        self._set_channels = [1, 32, 64, 128]
        self.batch_frame = None

        self.set_layer1_1 = SetBlock(self._set_channels[0], self._set_channels[1], 3, padding=1)
        self.set_layer1_2 = SetBlock(self._set_channels[1], self._set_channels[1], 3, True , padding=1)
        self.set_layer1_3 = SetBlock(self._set_channels[1], self._set_channels[2], 3, padding=1)
        self.set_layer1_4 = SetBlock(self._set_channels[2], self._set_channels[2], 3, True, padding=1)
        self.set_layer1_5 = SetBlock(self._set_channels[2], self._set_channels[3], 3, padding=1)
        self.set_layer1_6 = SetBlock(self._set_channels[3], self._set_channels[3], 3, padding=1)
        self.gl_layer1_1 = PipeBlock(self._set_channels[1], self._set_channels[2], 3, padding=1)
        self.gl_layer1_2 = PipeBlock(self._set_channels[2], self._set_channels[2], 3, True , padding=1)
        self.gl_layer1_3 = PipeBlock(self._set_channels[2], self._set_channels[3], 3, padding=1)
        self.gl_layer1_4 = PipeBlock(self._set_channels[3], self._set_channels[3], 3, padding=1)
        
        self.set_layer2_1 = SetBlock(self._set_channels[0], self._set_channels[1], 3, padding=1)
        self.set_layer2_2 = SetBlock(self._set_channels[1], self._set_channels[1], 3, True , padding=1)
        self.set_layer2_3 = SetBlock(self._set_channels[1], self._set_channels[2], 3, padding=1)
        self.set_layer2_4 = SetBlock(self._set_channels[2], self._set_channels[2], 3, True, padding=1)
        self.set_layer2_5 = SetBlock(self._set_channels[2], self._set_channels[3], 3, padding=1)
        self.set_layer2_6 = SetBlock(self._set_channels[3], self._set_channels[3], 3, padding=1)
        self.gl_layer2_1 = PipeBlock(self._set_channels[1], self._set_channels[2], 3, padding=1)
        self.gl_layer2_2 = PipeBlock(self._set_channels[2], self._set_channels[2], 3, True , padding=1)
        self.gl_layer2_3 = PipeBlock(self._set_channels[2], self._set_channels[3], 3, padding=1)
        self.gl_layer2_4 = PipeBlock(self._set_channels[3], self._set_channels[3], 3, padding=1)
        
        self.bin_num1 = [1, 3]
        self.bin_num2 = [1, 2, 3, 6]
        self.fc_bin1 = nn.ParameterList([nn.Parameter(nn.init.xavier_uniform_(torch.zeros(sum(self.bin_num1) * 2, 128, hidden_dim)))])
        self.fc_bin2 = nn.ParameterList([nn.Parameter(nn.init.xavier_uniform_(torch.zeros(sum(self.bin_num2) * 2, 128, hidden_dim)))])

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.xavier_uniform_(m.weight.data)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight.data)
                nn.init.constant_(m.bias.data, 0.0)

    def frame_max(self, x):
        if self.batch_frame is None:
            return torch.max(x, 1)
        else:
            _tmp = [torch.max(x[:, self.batch_frame[i]:self.batch_frame[i + 1], :, :, :], 1) for i in range(len(self.batch_frame) - 1)]
            max_list = torch.cat([_tmp[i][0] for i in range(len(_tmp))], 0)
            arg_max_list = torch.cat([_tmp[i][1] for i in range(len(_tmp))], 0)
            return max_list, arg_max_list

    def forward(self, silho, batch_frame=None):
        # n: batch_size, s: frame_num, k: keypoints_num, c: channel
        if batch_frame is not None:
            batch_frame = batch_frame[0].data.cpu().numpy().tolist()
            _ = len(batch_frame)
            for i in range(len(batch_frame)):
                if batch_frame[-(i + 1)] != 0:
                    break
                else:
                    _ -= 1
            batch_frame = batch_frame[:_]
            frame_sum = np.sum(batch_frame)
            if frame_sum < silho.size(1): silho = silho[:, :frame_sum, :, :]
            self.batch_frame = [0] + np.cumsum(batch_frame).tolist()
        n = silho.size(0)
        silho = silho.unsqueeze(2)

        feat1 = list()
        x = self.set_layer1_2(self.set_layer1_1(silho[:, :, :, :12, :]))
        gl = self.gl_layer1_2(self.gl_layer1_1(self.frame_max(x)[0]))
        x = self.set_layer1_4(self.set_layer1_3(x))
        gl = self.gl_layer1_4(self.gl_layer1_3(gl + self.frame_max(x)[0]))
        x = self.set_layer1_6(self.set_layer1_5(x))
        x1 = self.frame_max(x)[0]
        gl1 = gl + x1
        n, c, h, w = x1.size()
        for num_bin in self.bin_num1:
            z = x1.view(n, c, num_bin, -1)
            z = z.mean(3) + z.max(3)[0]
            feat1.append(z)
            z = gl1.view(n, c, num_bin, -1)
            z = z.mean(3) + z.max(3)[0]
            feat1.append(z)
        feat1 = torch.cat(feat1, 2).permute(2, 0, 1).contiguous()
        feat1 = feat1.matmul(self.fc_bin1[0])
        feat1 = feat1.permute(1, 0, 2).contiguous()

        feat2 = list()
        x = self.set_layer2_2(self.set_layer2_1(silho[:, :, :, 40:, :]))
        gl = self.gl_layer2_2(self.gl_layer2_1(self.frame_max(x)[0]))
        x = self.set_layer2_4(self.set_layer2_3(x))
        gl = self.gl_layer2_4(self.gl_layer2_3(gl + self.frame_max(x)[0]))
        x = self.set_layer2_6(self.set_layer2_5(x))
        x2 = self.frame_max(x)[0]
        gl2 = gl + x2
        n, c, h, w = x2.size()
        for num_bin in self.bin_num2:
            z = x2.view(n, c, num_bin, -1)
            z = z.mean(3) + z.max(3)[0]
            feat2.append(z)
            z = gl2.view(n, c, num_bin, -1)
            z = z.mean(3) + z.max(3)[0]
            feat2.append(z)
        feat2 = torch.cat(feat2, 2).permute(2, 0, 1).contiguous()
        feat2 = feat2.matmul(self.fc_bin2[0])
        feat2 = feat2.permute(1, 0, 2).contiguous()

        feat_id = torch.cat([feat1, feat2], dim=1)
        return feat_id
