import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision

import numpy as np


class BasicConv2d(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, pooling=False, **kwargs):
        super(BasicConv2d, self).__init__()
        self.pooling = pooling
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, bias=False, **kwargs)
        if pooling: self.pool2d = nn.MaxPool2d(2)
    def forward(self, x):
        n, s, c, h, w = x.size()
        x = x.view(n*s, c, h, w)
        x = F.leaky_relu(self.conv(x), inplace=True)
        if self.pooling: x = self.pool2d(x)
        _, c, h, w = x.size()
        x = x.view(n, s, c, h, w)
        return x


class TPNet(nn.Module):
    def __init__(self, hidden_dim):
        super(TPNet, self).__init__()
        self.hidden_dim = hidden_dim
        self.batch_frame = None

        self.channel_list = [1, 32, 64, 128]
        self.set_layer1 = BasicConv2d(self.channel_list[0], self.channel_list[1], 3, padding=1)
        self.set_layer2 = BasicConv2d(self.channel_list[1], self.channel_list[1], 3, True, padding=1)
        self.set_layer3 = BasicConv2d(self.channel_list[1], self.channel_list[2], 3, padding=1)
        self.set_layer4 = BasicConv2d(self.channel_list[2], self.channel_list[2], 3, True, padding=1)
        self.set_layer5 = BasicConv2d(self.channel_list[2], self.channel_list[3], 3, padding=1)
        self.set_layer6 = BasicConv2d(self.channel_list[3], self.channel_list[3], 3, True, padding=1)

        self.lsp1 = nn.MaxPool3d(kernel_size=(3, 1, 1), stride=(1, 1, 1))
        self.lsp2 = nn.MaxPool3d(kernel_size=(5, 1, 1), stride=(2, 1, 1))

        self.bin_num = [1, 2, 3, 6]
        self.fc_bin = nn.ParameterList([nn.Parameter(nn.init.xavier_uniform_(torch.zeros(sum(self.bin_num) * 2, 128, self.hidden_dim)))])

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.xavier_uniform_(m.weight.data)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight.data)
                nn.init.constant_(m.bias.data, 0.0)

    def frame_max(self, x):
        if self.batch_frame is None:
            return torch.max(x, 1)
        else:
            _tmp = [torch.max(x[:, self.batch_frame[i]:self.batch_frame[i + 1], :, :, :], 1) for i in range(len(self.batch_frame) - 1)]
            max_list = torch.cat([_tmp[i][0] for i in range(len(_tmp))], 0)
            arg_max_list = torch.cat([_tmp[i][1] for i in range(len(_tmp))], 0)
            return max_list, arg_max_list

    def forward(self, silho, batch_frame=None):
        # n: batch_size, s: frame_num, k: keypoints_num, c: channel
        if batch_frame is not None:
            batch_frame = batch_frame[0].data.cpu().numpy().tolist()
            _ = len(batch_frame)
            for i in range(len(batch_frame)):
                if batch_frame[-(i + 1)] != 0:
                    break
                else:
                    _ -= 1
            batch_frame = batch_frame[:_]
            frame_sum = np.sum(batch_frame)
            if frame_sum < silho.size(1): silho = silho[:, :frame_sum, :, :]
            self.batch_frame = [0] + np.cumsum(batch_frame).tolist()
        n = silho.size(0)
        x = silho.unsqueeze(2)
        while x.size(1) < 5: x = torch.cat([x, x], dim=1)
        del silho
        x = x[:, :, :, 40:, :]

        x = self.set_layer2(self.set_layer1(x))
        x = self.set_layer4(self.set_layer3(x))
        x = self.set_layer6(self.set_layer5(x))
        
        x1 = self.lsp1(x.permute(0, 2, 1, 3, 4).contiguous())
        x2 = self.lsp2(x.permute(0, 2, 1, 3, 4).contiguous())
        feat_map1 = self.frame_max(x1.permute(0, 2, 1, 3, 4).contiguous())[0]
        feat_map2 = self.frame_max(x2.permute(0, 2, 1, 3, 4).contiguous())[0]

        feat = list()
        bs, nc, _, _ = feat_map1.size()
        for num_bin in self.bin_num:
            z = feat_map1.view(bs, nc, num_bin, -1)
            z = z.mean(3) + z.max(3)[0]
            feat.append(z)
            z = feat_map2.view(bs, nc, num_bin, -1)
            z = z.mean(3) + z.max(3)[0]
            feat.append(z)
        feat = torch.cat(feat, 2).permute(2, 0, 1).contiguous()
        feat = feat.matmul(self.fc_bin[0])
        feat = feat.permute(1, 0, 2).contiguous()

        return feat
