import os
import cv2
import pickle
import numpy as np
import xarray as xr
import torch.utils.data as tordata


class DataSet(tordata.Dataset):
    def __init__(self, seqs_dir, seqs_label, seqs_type, seqs_view, cache, resolution, docut_padding=False):
        self.seqs_dir = seqs_dir
        self.seqs_label = seqs_label
        self.seqs_type = seqs_type
        self.seqs_view = seqs_view
        self.cache = cache
        self.resolution = resolution
        self.docut_padding = docut_padding
        self.data_size = len(self.seqs_label)
        self.data_fast = [None] * self.data_size
        self.data_slow = [None] * self.data_size
        self.frameset_fast = [None] * self.data_size
        self.frameset_slow = [None] * self.data_size

        self.seqs_label_set = set(self.seqs_label)
        self.seqs_type_set = set(self.seqs_type)
        self.seqs_view_set = set(self.seqs_view)
        _ = np.zeros((len(self.seqs_label_set), len(self.seqs_type_set), len(self.seqs_view_set))).astype('int')
        _ -= 1
        self.index_dict = xr.DataArray(
            _,
            coords={'seqs_label': sorted(list(self.seqs_label_set)), 'seqs_type': sorted(list(self.seqs_type_set)), 'seqs_view': sorted(list(self.seqs_view_set))},
            dims=['seqs_label', 'seqs_type', 'seqs_view'])
        for i in range(self.data_size):
            _seqs_label, _seqs_type, _seqs_view = self.seqs_label[i], self.seqs_type[i], self.seqs_view[i]
            self.index_dict.loc[_seqs_label, _seqs_type, _seqs_view] = i

    def __len__(self):
        return len(self.seqs_label)

    def __getitem__(self, index):
        if not self.cache:
            data_fast = [self.__loadimg__(dirpath, self.resolution[0]) for dirpath in self.seqs_dir[index]]
            frameset_fast = [set(feature.coords['frame'].values.tolist()) for feature in data_fast]
            frameset_fast = list(set.intersection(*frameset_fast))
            data_slow = [self.__loadimg__(dirpath, self.resolution[1]) for dirpath in self.seqs_dir[index]]
            frameset_slow = [set(feature.coords['frame'].values.tolist()) for feature in data_slow]
            frameset_slow = list(set.intersection(*frameset_slow))
        elif self.data_fast[index] is None and self.data_slow[index] is None:
            data_fast = [self.__loadimg__(dirpath, self.resolution[0]) for dirpath in self.seqs_dir[index]]
            frameset_fast = [set(feature.coords['frame'].values.tolist()) for feature in data_fast]
            frameset_fast = list(set.intersection(*frameset_fast))
            self.data_fast[index] = data_fast
            self.frameset_fast[index] = frameset_fast
            data_slow = [self.__loadimg__(dirpath, self.resolution[1]) for dirpath in self.seqs_dir[index]]
            frameset_slow = [set(feature.coords['frame'].values.tolist()) for feature in data_slow]
            frameset_slow = list(set.intersection(*frameset_slow))
            self.data_slow[index] = data_slow
            self.frameset_slow[index] = frameset_slow
        else:
            data_fast = self.data_fast[index]
            frameset_fast = self.frameset_fast[index]
            data_slow = self.data_slow[index]
            frameset_slow = self.frameset_slow[index]
        return data_fast, data_slow, frameset_fast, frameset_slow, self.seqs_view[index], self.seqs_type[index], self.seqs_label[index]

    def __loadimg__(self, dirpath, resolution):
        if self.docut_padding:
            self.cut_padding = int(float(resolution)/64*10)
            return self.img2xarray(dirpath, resolution)[:, :, self.cut_padding:-self.cut_padding].astype('float32') / 255.0
        else:
            return self.img2xarray(dirpath, resolution).astype('float32') / 255.0

    def img2xarray(self, dirpath, resolution):
        frame_list = list()
        filenames = sorted(list(os.listdir(dirpath)))
        for filename in filenames:
            img = cv2.imread(os.path.join(dirpath, filename))
            img = cv2.resize(img, (resolution, resolution))
            img = np.reshape(img, [resolution, resolution, -1])[:, :, 0]
            frame_list.append(img)
        num_list = list(range(len(frame_list)))
        data_dict = xr.DataArray(frame_list, coords={'frame': num_list}, dims=['frame', 'img_y', 'img_x'],)
        return data_dict

    def load_all_data(self):
        for i in range(self.data_size):
            self.__getitem__(i)
