import argparse
import os
import shutil
import itertools
import numpy as np
import xarray as xr
from datetime import datetime

import torch
import torch.nn.functional as F

from config import conf as config
from initialization import initialization

times = 1
num_rank = 5


# Compute euclidean distance
def cuda_dist(x, y):
    x = torch.from_numpy(x).cuda()
    y = torch.from_numpy(y).cuda()
    dist = torch.sum(x ** 2, 1).unsqueeze(1) + torch.sum(y ** 2, 1).unsqueeze(1).transpose(0, 1) - 2 * torch.matmul(x, y.transpose(0, 1))
    dist = torch.sqrt(F.relu(dist))
    return dist
    
def evaluation(data, config):
    list_gfeat, list_gview, list_gtype, list_glabel = data[0]
    list_pfeat, list_pview, list_ptype, list_plabel = data[1]

    list_glabel, list_plabel = np.array(list_glabel), np.array(list_plabel)
    set_ptype = sorted(list(set(list_ptype)))
    acc = np.zeros([len(set_ptype), num_rank])
    for p, ptype in enumerate(set_ptype):
        gallery_x, gallery_y = list_gfeat, list_glabel
        pseq_mask = np.isin(list_ptype, [ptype])
        probe_x, probe_y = list_pfeat[pseq_mask, :], list_plabel[pseq_mask]
        dist = cuda_dist(probe_x, gallery_x)
        idx = dist.sort(1)[1].cpu().numpy()
        acc[p, :] = np.round(np.sum(np.cumsum(np.reshape(probe_y, [-1, 1]) == gallery_y[idx[:, 0:num_rank]], 1) > 0, 0) * 100 / dist.shape[0], 2)
        #print(p, ptype, dist.shape, gallery_x.shape, gallery_y.shape, probe_x.shape, probe_y.shape, acc[p, 0])
    return acc, set_ptype


for _time in range(3):
    # Train
    print(' ==> Training %d START' % _time)
    is_train = True
    parser = argparse.ArgumentParser(description='Train')
    parser.add_argument('--cache', default=False, type=bool, help='cache: if set as TRUE all the training data will be loaded at once before the training start. Default: TRUE')
    opt = parser.parse_args()
    model = initialization(config, cache=opt.cache, is_train=is_train)
    model.fit()
    print(' ==> Training COMPLETE\n')

    # Test
    print(' ==> Testing %d START' % _time)
    is_train = False
    parser = argparse.ArgumentParser(description='Test')
    parser.add_argument('--batch_size', default='1', type=int, help='batch_size: batch size for parallel test. Default: 1')
    parser.add_argument('--cache', default=False, type=bool, help='cache: if set as TRUE all the test data will be loaded at once before the transforming start. Default: FALSE')
    opt = parser.parse_args()
    model = initialization(config, cache=opt.cache, is_train=is_train)
    best_iter, best_acc = 0, 0
    for iter in range(config['model']['save_iter'], config['model']['total_iter']+1, config['model']['save_iter']):
        print(' ==> Loading the model of iteration %d...' % iter)
        np.set_printoptions(precision=2, floatmode='fixed')
        model.load(iter)
        time = datetime.now()
        for i in range(times):
            data = model.transform(opt.batch_size)
            acc_, list_ptype = evaluation(data, config['data'])
            acc = acc_ if i == 0 else acc+acc_
        acc = acc / times           
        print('Evaluation complete. Cost:', datetime.now() - time)
        print('***** Rank-1 of each type *****')
        for p, ptype in enumerate(list_ptype): print('      GALLERY all v.s. PROBE %s: %.3f' % (ptype, acc[p, 0]))
        if np.mean(acc[:, 0]) >= best_acc: best_iter, best_acc = iter, np.mean(acc[:, 0])
        print('***** best_iter: %d\t best_acc: %0.3f\t ave_acc: %0.3f' % (best_iter, best_acc, np.mean(acc[:, 0])))
    print(' ==> Testing FINISH\n')

    # Save
    print(' ==> Recoding')
    print(' ==> Loading the model of best iteration %d...' % best_iter)
    model.load(best_iter)
    time = datetime.now()
    np.set_printoptions(precision=2, floatmode='fixed')
    time = datetime.now()
    for i in range(times):
        data = model.transform(opt.batch_size)
        acc_, list_ptype = evaluation(data, config['data'])
        acc = acc_ if i == 0 else acc+acc_
    acc = acc / times           
    print('Evaluation complete. Cost:', datetime.now() - time)
    print('***** Rank-1 of each type *****')
    for p, ptype in enumerate(list_ptype): print('      GALLERY all v.s. PROBE %s: %.3f' % (ptype, acc[p, 0]))
    with open('RECORD', 'a') as fout:
        fout.write('model_name: {}\t batch_size: {}\t frame_num: {}\n'.format(config['model']['model_name'], config['model']['batch_size'], config['model']['frame_num']))
        fout.write('Rank-1 (Treadmill-B):\n')      
        for p, ptype in enumerate(list_ptype): fout.write('      GALLERY all v.s. PROBE %s: %.3f\n' % ((ptype, acc[p, 0])))		
    print(' ==> Recoding FINISH\n\n')
